class NotSupportOperatorException(Exception):
    def __init__(self, operator: str):
        super().__init__(f"not support operator {operator}")


class NotSupportRelationshipQueryException(Exception):
    def __init__(self, operator: str):
        super().__init__(f"only one-to-many, many-to-many relationship fields support {operator}")


class InvalidFieldException(Exception):
    def __init__(self, field: str):
        super().__init__(f"invalid field name {field}")

class NotFoundException(Exception):
    pass


from enum import Enum


class RoutesEnum(str, Enum):
    get_many = 'get_many'
    get_one = 'get_one'
    create_one = 'create_one',
    create_many = 'create_many',
    update_one = 'update_one'
    update_many = 'update_many'
    delete_many = 'delete_many'


class CrudActions(str, Enum):
    read_all = 'read',
    read_one = 'read',
    create_one = 'create',
    create_many = 'create',
    update_one = 'update',
    update_many = 'update',
    delete_many = 'delete'


class QuerySortType(str, Enum):
    ASC = 'ASC'
    DESC = 'DESC'



import inspect
from typing import (
    Any,
    Callable,
    List,
    Type,
    TypeVar,
    Annotated,
    cast,
    Union,
    Dict
)
from functools import wraps
from fastapi import (
    APIRouter,
    status,
    Body,
    Depends,
    Request,
    Path,
    Query,
    HTTPException,
    BackgroundTasks
)
from .enums import RoutesEnum
from .models import (
    CrudOptions,
    AbstractResponseModel,
    RouteOptions,
    JoinOptions
)
from .types import QuerySortDict, CreateSchemaType, UpdateSchemaType
from .config import BetterCrudGlobalConfig, RoutesSchema
from .helper import get_serialize_model, get_route_summary
from .depends import (
    CrudAction,
    StateAction,
    GetQuerySearch,
    GetQueryLoads,
    GetQuerySorts,
    GetQueryJoins,
)
from fastapi_pagination import pagination_ctx
from fastapi_pagination.bases import AbstractPage
from .exceptions import NotFoundException

T = TypeVar("T")
CRUD_CLASS_KEY = "__crud_class__"
UNBIND_KIND_TYPE = (
    inspect.Parameter.VAR_POSITIONAL,
    inspect.Parameter.VAR_KEYWORD
)
INCLUDE_DELETED_KEY = "include_deleted"


def crud_routes_factory(router: APIRouter, cls: Type[T], options: CrudOptions) -> Type[T]:
    create_schema_type = cast(CreateSchemaType, options.dto.create)
    update_schema_type = cast(UpdateSchemaType, options.dto.update)
    page_schema_type = cast(AbstractPage, BetterCrudGlobalConfig.page_schema)
    response_schema_type = cast(
        AbstractResponseModel,
        BetterCrudGlobalConfig.response_schema
    )

    serialize = options.serialize

    async def get_many(
        self,
        request: Request,
        search: Dict = Depends(
            GetQuerySearch(options.query.filter)
        ),
        joins: JoinOptions = Depends(
            GetQueryJoins(options.query.joins)
        ),
        sorts: List[QuerySortDict] = Depends(
            GetQuerySorts(options.query.sort)),
    ):
        return await self.service.crud_get_many(
            request=request,
            joins=joins,
            search=search,
            sorts=sorts,
            soft_delete=options.query.soft_delete,
            include_deleted=request.query_params.get(
                INCLUDE_DELETED_KEY) == "true" if options.query.allow_include_deleted else False
        )

    async def get_one(
        self,
        request: Request,
        joins: JoinOptions = Depends(
            GetQueryLoads(options.query.joins)
        ),
        id: Union[int, str] = Path(..., title="The ID of the item to get")
    ):
        try:
            return await self.service.crud_get_one(
                request,
                id,
                joins=joins
            )
        except NotFoundException:
            raise HTTPException(
                status.HTTP_404_NOT_FOUND,
                detail="No data found"
            )

    async def create_one(
        self,
        model: Annotated[create_schema_type, Body()],  # type: ignore
        request: Request,
        background_tasks: BackgroundTasks
    ):
        return await self.service.crud_create_one(
            request,
            model,
            background_tasks=background_tasks
        )

    async def create_many(
        self,
        model: Annotated[List[create_schema_type], Body()],  # type: ignore
        request: Request,
        background_tasks: BackgroundTasks
    ):
        return await self.service.crud_create_many(
            request,
            model,
            background_tasks=background_tasks
        )

    async def update_one(
        self,
        model: Annotated[update_schema_type, Body()],  # type: ignore
        request: Request,
        background_tasks: BackgroundTasks,
        id: Union[int, str] = Path(..., title="The ID of the item to get")
    ):
        try:
            return await self.service.crud_update_one(
                request,
                id,
                model,
                background_tasks=background_tasks
            )
        except NotFoundException:
            raise HTTPException(
                status.HTTP_404_NOT_FOUND,
                detail="No data found"
            )

    async def update_many(
        self,
        request: Request,
        models: Annotated[List[update_schema_type], Body()],  # type: ignore
        background_tasks: BackgroundTasks,
        ids: str = Path(...,
                        description="Primary key values, use commas to separate multiple values")
    ):
        id_list = ids.split(",")
        if len(id_list) != len(models):
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                detail="The id and payload length do not match"
            )
        try:
            return await self.service.crud_update_many(
                request,
                id_list,
                models,
                background_tasks=background_tasks
            )
        except NotFoundException:
            raise HTTPException(
                status.HTTP_404_NOT_FOUND,
                detail="No data found"
            )

    async def delete_many(
        self,
        request: Request,
        background_tasks: BackgroundTasks,
        ids: str = Path(...,
                        description="Primary key values, use commas to separate multiple values")
    ):
        id_list = ids.split(",")
        try:
            return await self.service.crud_delete_many(
                request,
                id_list,
                soft_delete=options.query.soft_delete,
                background_tasks=background_tasks
            )
        except NotFoundException:
            raise HTTPException(
                status.HTTP_404_NOT_FOUND,
                detail="No data found"
            )

    cls.get_many = get_many
    cls.create_one = create_one
    cls.create_many = create_many
    cls.update_one = update_one
    cls.update_many = update_many
    cls.delete_many = delete_many
    cls.get_one = get_one

    function_members = inspect.getmembers(cls, inspect.isfunction)
    functions_set = set(func for _, func in function_members)
    for func in functions_set:
        _update_route_endpoint_signature(cls, func, options)

    for schema in RoutesSchema:
        router_name = schema["name"].value
        path = schema["path"]
        method = schema["method"]
        if options.routes and options.routes.only is not None:
            if router_name not in options.routes.only:
                continue
        if options.routes and options.routes.exclude:
            if router_name in options.routes.exclude:
                continue
        overrides = list(filter(lambda route: route.path ==
                         path and method in route.methods, router.routes))
        if overrides:
            continue
        endpoint = getattr(cls, router_name)

        def decorator(func: Callable) -> Callable:
            @wraps(func)
            async def wrapper(*args: Any, **kwargs: Any) -> Any:
                if options.params:
                    for key in options.params.keys():
                        kwargs.pop(key)
                if INCLUDE_DELETED_KEY in kwargs:
                    kwargs.pop(INCLUDE_DELETED_KEY)
                endpoint_output = await func(*args, **kwargs)
                if response_schema_type:
                    return response_schema_type.create(endpoint_output)
                return endpoint_output
            return wrapper
        endpoint_wrapper = decorator(endpoint)
        response_model = get_serialize_model(serialize, router_name)
        if router_name == RoutesEnum.get_many:
            response_model = Union[
                page_schema_type[response_model],
                List[response_model]
            ]
        elif router_name in [RoutesEnum.create_many, RoutesEnum.update_many, RoutesEnum.delete_many]:
            response_model = List[response_model]

        if response_schema_type:
            response_model = response_schema_type[response_model]

        dependencies = None
        route_options: RouteOptions = getattr(
            options.routes,
            router_name,
            None
        )
        if route_options and route_options.dependencies is not None:
            dependencies = [*route_options.dependencies]
        if dependencies is None and options.routes.dependencies:
            dependencies = [*options.routes.dependencies]

        if dependencies is None:
            dependencies = []
        if router_name == RoutesEnum.get_many:
            dependencies.append(
                Depends(
                    pagination_ctx(BetterCrudGlobalConfig.page_schema)
                )
            )
        router.add_api_route(
            path,
            endpoint_wrapper,
            methods=[method],
            summary=get_route_summary(route_options, options.summary_vars),
            dependencies=[
                Depends(CrudAction(
                    options.feature,
                    route_options.action if route_options else None,
                    BetterCrudGlobalConfig.action_map,
                    router_name
                )),
                *dependencies,
                Depends(StateAction(options.auth, options.params)),
            ],
            response_model=response_model,
        )
    return cls


def _update_route_endpoint_signature(
    cls: Type[Any],
    endpoint: Callable,
    options: CrudOptions
) -> None:
    old_signature = inspect.signature(endpoint)
    old_parameters: List[inspect.Parameter] = list(
        old_signature.parameters.values())
    old_first_parameter = old_parameters[0]
    new_first_parameter = old_first_parameter.replace(default=Depends(cls))
    new_parameters = [new_first_parameter] + [
        parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY)
        for parameter in old_parameters[1:]
    ]
    if options.params:
        for key, param in options.params.items():
            new_param = inspect.Parameter(
                key,
                inspect.Parameter.KEYWORD_ONLY,
                annotation=Annotated[
                    int if param.type == "int" else str, Path(title="")
                ]
            )
            new_parameters.append(new_param)
    if endpoint == cls.get_many:
        if options.query.allow_include_deleted:
            new_param = inspect.Parameter(
                INCLUDE_DELETED_KEY,
                inspect.Parameter.KEYWORD_ONLY,
                annotation=Annotated[bool, Query(
                    description="include deleted data")]
            )
            new_parameters.append(new_param)

    new_signature = old_signature.replace(parameters=new_parameters)
    setattr(endpoint, "__signature__", new_signature)



from typing import Dict, Optional, ClassVar, List
from importlib import import_module
from .models import (
    GlobalQueryOptions,
    RoutesModel,
    QueryDelimOptions,
    AbstractResponseModel,
    BackendConfigModel
)
from .types import (
    GlobalQueryOptionsDict,
    RoutesModelDict,
    QueryDelimOptionsDict,
    BackendConfigDict,
    RouteSchema
)
from .pagination import Page
from fastapi_pagination.bases import AbstractPage
from .enums import RoutesEnum, CrudActions

DEFAULT_SOFT_DELETED_FIELD_KEY = "deleted_at"

DEFAULT_ACTION_MAP = {
    RoutesEnum.get_many: CrudActions.read_all,
    RoutesEnum.get_one: CrudActions.read_one,
    RoutesEnum.create_one: CrudActions.create_one,
    RoutesEnum.create_many: CrudActions.create_many,
    RoutesEnum.update_one: CrudActions.update_one,
    RoutesEnum.update_many: CrudActions.update_many,
    RoutesEnum.delete_many: CrudActions.delete_many
}


RoutesSchema: List[RouteSchema] = [
    {
        "name": RoutesEnum.get_many,
        "path": '',
        "method": "GET"
    },
    {
        "name": RoutesEnum.create_one,
        "path": '',
        "method": "POST"
    },
    {
        "name": RoutesEnum.create_many,
        "path": '/bulk',
        "method": "POST"
    },
    {
        "name": RoutesEnum.get_one,
        "path": '/{id}',
        "method": "GET"
    },
    {
        "name": RoutesEnum.update_one,
        "path": '/{id}',
        "method": "PUT"
    },
    {
        "name": RoutesEnum.update_many,
        "path": '/{ids}/bulk',
        "method": "PUT"
    },
    {
        "name": RoutesEnum.delete_many,
        "path": '/{ids}',
        "method": "DELETE"
    }
]


class BetterCrudGlobalConfig:
    query: ClassVar[GlobalQueryOptions] = GlobalQueryOptions()
    routes: ClassVar[Optional[RoutesModel]] = RoutesModel()
    delim_config: ClassVar[Optional[QueryDelimOptions]] = None
    soft_deleted_field_key: ClassVar[Optional[str]
                                     ] = DEFAULT_SOFT_DELETED_FIELD_KEY
    action_map: ClassVar[Optional[Dict[str, str]]] = None
    page_schema: ClassVar[Optional[AbstractPage]] = Page
    response_schema: ClassVar[Optional[AbstractResponseModel]] = None
    backend_config: ClassVar[BackendConfigModel] = None

    @classmethod
    def init(
        cls,
        *,
        backend_config: Optional[BackendConfigDict] = None,
        query: Optional[GlobalQueryOptionsDict] = {},
        routes: Optional[RoutesModelDict] = {},
        delim_config: Optional[QueryDelimOptionsDict] = {},
        soft_deleted_field_key: Optional[str] = None,
        action_map: Optional[Dict[str, str]] = None,
        page_schema: Optional[AbstractPage] = Page,
        response_schema: Optional[AbstractResponseModel] = None
    ) -> None:
        cls.query = GlobalQueryOptions(**query)
        cls.routes = RoutesModel(**routes)
        cls.delim_config = QueryDelimOptions(**delim_config)
        if soft_deleted_field_key:
            cls.soft_deleted_field_key = soft_deleted_field_key
        cls.page_schema = page_schema
        cls.response_schema = response_schema
        cls.action_map = action_map or DEFAULT_ACTION_MAP
        cls.backend_config = BackendConfigModel(**backend_config)
        if cls.backend_config.backend != "custom":
            import_module(
                f".service.{cls.backend_config.backend}", package="better_crud")



__version__ = "0.1.4"
__title__ = "better_crud"
__author__ = "bigrivi"
__author_email__ = "sunjianghong@gmail.com"
__copyright__ = "Copyright 2024 bigrivi"
__description__ = "A better CRUD library for FastAPI."
__url__ = "https://github.com/bigrivi/better_crud"
__license__ = "Apache License 2.0"



from typing import (
    List,
    Optional,
    Union,
    Any,
    Dict,
    Callable,
    Sequence,
    TypeVar,
    Type,
    Literal
)
from abc import ABC, abstractmethod
from pydantic import BaseModel, Field, ConfigDict
from .enums import RoutesEnum, QuerySortType
from .types import DBSessionFactory, BackendType
C = TypeVar("C")


class DtoModel(BaseModel):
    model_config = ConfigDict(extra="forbid")
    create: Optional[Any] = None
    update: Optional[Any] = None


class AuthModel(BaseModel):
    model_config = ConfigDict(extra="forbid")
    filter_: Callable[[Any], Dict] = Field(default=None, alias='filter')
    persist: Callable[[Any], Dict] = None


class SerializeModel(BaseModel):
    model_config = ConfigDict(extra="forbid")
    base: Any = Field(...)
    get_many: Optional[Any] = None
    get_one: Optional[Any] = None
    create_one: Optional[Any] = None
    create_many: Optional[Any] = None
    update_one: Optional[Any] = None
    update_many: Optional[Any] = None
    delete_many: Optional[Any] = None


class RouteOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")
    dependencies: Optional[Sequence[Any]] = None
    summary: Optional[str] = None
    action: Optional[str] = None


class RoutesModel(BaseModel):
    model_config = ConfigDict(extra="forbid")
    dependencies: Optional[Sequence[Any]] = None
    only: Optional[List[RoutesEnum]] = None
    exclude: Optional[List[RoutesEnum]] = None
    get_many: Optional[RouteOptions] = None
    get_one: Optional[RouteOptions] = None
    create_one: Optional[RouteOptions] = None
    create_many: Optional[RouteOptions] = None
    update_one: Optional[RouteOptions] = None
    update_many: Optional[RouteOptions] = None
    delete_many: Optional[RouteOptions] = None


class QuerySortModel(BaseModel):
    field: str
    sort: QuerySortType


class JoinOptionModel(BaseModel):
    select: Optional[bool] = True
    join: Optional[bool] = True
    select_only_detail: Optional[bool] = False
    additional_filter_fn: Optional[Callable[[Any], List[Any]]] = None
    alias: Any = None


JoinOptions = Dict[str, JoinOptionModel]


class QueryOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")
    joins: Optional[JoinOptions] = None
    soft_delete: Optional[bool] = None
    allow_include_deleted: Optional[bool] = False
    filter: Union[Optional[Dict], Callable[[Any], Dict]] = None
    sort: Optional[List[QuerySortModel]] = None


class PathParamModel(BaseModel):
    field: str
    type: Literal["str", "int"]


class CrudOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")
    dto: DtoModel = None
    serialize: Optional[SerializeModel] = None
    routes: Optional[RoutesModel] = None
    feature: str = None
    query: Optional[QueryOptions] = None
    auth: Optional[AuthModel] = None
    summary_vars: Dict = None
    params: Optional[Dict[str, PathParamModel]] = None


class GlobalQueryOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")
    soft_delete: Optional[bool] = False
    sort: Optional[List[QuerySortModel]] = None
    allow_include_deleted: Optional[bool] = False


class QueryDelimOptions(BaseModel):
    model_config = ConfigDict(extra="forbid")
    delim: Optional[str] = "||"
    delim_str: Optional[str] = ","


class AbstractResponseModel(BaseModel, ABC):
    @classmethod
    @abstractmethod
    def create(
        cls: Type[C],
        *args: Any,
        **kwargs: Any,
    ) -> C:
        raise NotImplementedError


class SqlalchemyBackendModel(BaseModel):
    db_session: DBSessionFactory


class BackendConfigModel(BaseModel):
    backend: BackendType = "sqlalchemy"
    sqlalchemy: SqlalchemyBackendModel



from typing import (
    Any,
    Dict,
    List,
    Union,
    TypeVar,
    Generic,
    Optional,
    Sequence
)
from datetime import datetime
import functools
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import MANYTOMANY, MANYTOONE, ONETOMANY, noload, joinedload
from sqlalchemy.sql.selectable import Select
from sqlalchemy import or_, update, delete, and_, func, select
from sqlalchemy.orm.interfaces import ORMOption
from fastapi import Request, BackgroundTasks
from fastapi_pagination.ext.sqlalchemy import paginate
from fastapi_pagination.bases import AbstractPage
from ...helper import decide_should_paginate, build_join_options_tree
from ..abstract import AbstractCrudService
from ...types import QuerySortDict, ID_TYPE, CreateSchemaType, UpdateSchemaType
from ...models import JoinOptions, JoinOptionModel
from ...backend import register_backend

from ...config import BetterCrudGlobalConfig
from .helper import (
    create_many_to_many_instances,
    create_one_to_many_instances,
    create_many_to_one_instance,
    create_one_to_one_instance,
    inject_db_session,
    Provide
)
from ...exceptions import (
    NotSupportOperatorException,
    InvalidFieldException,
    NotSupportRelationshipQueryException,
    NotFoundException
)

ModelType = TypeVar("ModelType")
Selectable = TypeVar("Selectable", bound=Select[Any])


LOGICAL_OPERATOR_AND = "$and"
LOGICAL_OPERATOR_OR = "$or"


@register_backend("sqlalchemy")
class SqlalchemyCrudService(
    Generic[ModelType],
    AbstractCrudService[ModelType]
):

    entity: object = NotImplementedError

    def __init__(
        self,
        entity: object
    ):
        self.entity = entity
        self.primary_key = entity.__mapper__.primary_key[0].name
        self.entity_has_delete_column = hasattr(
            self.entity, BetterCrudGlobalConfig.soft_deleted_field_key)

    def prepare_order(
        self,
        query,
        sorts: List[QuerySortDict],
        joins: Optional[JoinOptions] = None
    ):
        order_bys = []
        if sorts:
            for sort_item in sorts:
                field = self.get_model_field(sort_item["field"], joins)
                sort = sort_item["sort"]
                if sort == "ASC":
                    order_bys.append(field.asc())
                elif sort == "DESC":
                    order_bys.append(field.desc())
        query = query.order_by(*order_bys)
        return query

    def create_search_field_object_condition(
        self,
        logical_operator: str,
        field: str,
        obj: Dict[str, Any],
        joins: Optional[JoinOptions] = None,
    ):
        logical_operator = logical_operator or LOGICAL_OPERATOR_AND
        if not isinstance(obj, dict):
            return None
        model_field = self.get_model_field(field, joins)
        keys = list(obj.keys())
        if len(keys) == 1:
            if keys[0] == LOGICAL_OPERATOR_OR:
                return self.create_search_field_object_condition(
                    LOGICAL_OPERATOR_OR,
                    field,
                    obj.get(LOGICAL_OPERATOR_OR),
                    joins
                )
            else:
                return self.build_query_expression(
                    model_field,
                    keys[0],
                    obj[keys[0]]
                )
        else:
            if logical_operator == LOGICAL_OPERATOR_OR:
                return or_(*(self.build_query_expression(
                    model_field,
                    operator,
                    obj[operator]
                ) for operator in keys))
            elif logical_operator == LOGICAL_OPERATOR_AND:
                clauses = []
                for operator in keys:
                    is_or = operator == LOGICAL_OPERATOR_OR
                    if isinstance(obj[operator], dict) and is_or:
                        clauses.append(
                            self.create_search_field_object_condition(
                                LOGICAL_OPERATOR_OR,
                                field,
                                obj.get(operator),
                                joins
                            )
                        )
                    else:
                        clauses.append(self.build_query_expression(
                            model_field, operator, obj[operator]))
                return and_(*clauses)

    def create_search_condition(
        self,
        search: Dict,
        joins: Optional[JoinOptions] = None,
    ) -> List[Any]:
        if not isinstance(search, dict) or not search:
            return []
        conds = []
        if LOGICAL_OPERATOR_AND in search:  # {$and: [...], ...}
            and_values = search.get(LOGICAL_OPERATOR_AND)
            if len(and_values) == 1:  # {$and: [{}]}
                conds.append(
                    and_(*self.create_search_condition(and_values[0], joins)))
            else:  # {$and: [{},{},...]}
                clauses = [
                    and_(*self.create_search_condition(and_value, joins))
                    for and_value in and_values
                ]
                conds.append(and_(*clauses))
        else:
            for field, value in search.items():
                field_is_or = field == LOGICAL_OPERATOR_OR
                if field_is_or and isinstance(value, list):
                    if len(value) == 1:
                        conds.append(
                            and_(
                                *self.create_search_condition(value[0], joins)
                            )
                        )
                    else:
                        clauses = [
                            and_(
                                *self.create_search_condition(or_value, joins)
                            ) for or_value in value
                        ]
                        conds.append(or_(*clauses))
                elif isinstance(value, Dict):
                    conds.append(
                        self.create_search_field_object_condition(
                            LOGICAL_OPERATOR_AND,
                            field,
                            value,
                            joins
                        )
                    )
                else:
                    conds.append(self.get_model_field(
                        field, joins) == value)
        return conds

    def _create_join_options(
        self,
        join_tree_nodes: List[Dict],
        request: Optional[Request] = None,
        from_detail: Optional[bool] = False
    ) -> Sequence[ORMOption]:
        options: Sequence[ORMOption] = []
        for join_tree_node in join_tree_nodes:
            field_key = join_tree_node["field_key"]
            config: JoinOptionModel = join_tree_node["config"]
            children: List[Dict] = join_tree_node["children"]
            join_field = self.get_model_field(field_key)
            should_select = config.select
            if not from_detail and config.select_only_detail:
                should_select = False
            if should_select:
                if config.additional_filter_fn:
                    filter_results = config.additional_filter_fn(request)
                    if not isinstance(filter_results, list):
                        filter_results = [filter_results]
                    loader = joinedload(join_field.and_(*filter_results))
                else:
                    loader = joinedload(join_field)
                if children:
                    options.append(loader.options(*self._create_join_options(
                        children,
                        request=request,
                        from_detail=from_detail
                    )))
                else:
                    options.append(loader)
            else:
                options.append(noload(join_field))
        return options

    def _build_query(
        self,
        search: Optional[Dict] = None,
        include_deleted: Optional[bool] = False,
        soft_delete: Optional[bool] = True,
        joins: Optional[JoinOptions] = None,
        sorts: List[QuerySortDict] = None,
        request: Optional[Request] = None,
        populate_existing: Optional[bool] = False
    ) -> Selectable:
        conds = []
        options = []
        if search:
            conds = conds + self.create_search_condition(search, joins)
        if self.entity_has_delete_column and soft_delete:
            soft_deleted_field = BetterCrudGlobalConfig.soft_deleted_field_key
            if not include_deleted:
                conds.append(or_(
                    getattr(self.entity, soft_deleted_field) > datetime.now(),
                    getattr(self.entity, soft_deleted_field).is_(None)
                ))
        stmt = select(self.entity)
        if joins:
            for field_key, config in joins.items():
                if config.join:
                    join_field = self.get_model_field(field_key)
                    if config.alias:
                        join_field = join_field.of_type(config.alias)
                    stmt = stmt.join(join_field, isouter=True)
            options = options + self._create_join_options(
                build_join_options_tree(joins),
                request=request
            )
        if options:
            stmt = stmt.options(*options)
        stmt = stmt.distinct()
        stmt = stmt.where(*conds)
        stmt = self.prepare_order(stmt, sorts, joins)
        if populate_existing:
            stmt = stmt.execution_options(populate_existing=populate_existing)
        return stmt

    @inject_db_session
    async def crud_get_many(
        self,
        request: Optional[Request] = None,
        search: Optional[Dict] = None,
        include_deleted: Optional[bool] = False,
        soft_delete: Optional[bool] = False,
        sorts: List[QuerySortDict] = None,
        joins: Optional[JoinOptions] = None,
        db_session: Optional[AsyncSession] = Provide(),
    ) -> Union[AbstractPage[ModelType], List[ModelType]]:
        query = self._build_query(
            search=search,
            include_deleted=include_deleted,
            soft_delete=soft_delete,
            joins=joins,
            sorts=sorts,
            request=request
        )
        if decide_should_paginate():
            return await paginate(db_session, query)
        result = await db_session.execute(query)
        return result.unique().scalars().all()

    @inject_db_session
    async def crud_get_one(
        self,
        request: Request,
        id: ID_TYPE,
        joins: Optional[JoinOptions] = None,
        db_session: Optional[AsyncSession] = Provide()
    ) -> ModelType:
        entity = await self._get(
            id,
            db_session,
            options=self._create_join_options(
                build_join_options_tree(joins),
                request=request,
                from_detail=True
            )
        )
        if not entity:
            raise NotFoundException()
        return entity

    @inject_db_session
    async def crud_create_one(
        self,
        request: Request,
        model: CreateSchemaType,
        background_tasks: Optional[BackgroundTasks] = None,
        db_session: Optional[AsyncSession] = Provide()
    ) -> ModelType:
        relationships = self.entity.__mapper__.relationships
        extra_data = await self.on_before_create(
            model,
            background_tasks=background_tasks
        )
        model_data: Dict = model.model_dump(exclude_unset=True)
        if extra_data:
            model_data.update(extra_data)
        if request:
            if hasattr(request.state, "auth_persist"):
                model_data.update(request.state.auth_persist)
            if hasattr(request.state, "params_filter"):
                model_data.update(request.state.params_filter)
        for key, value in model_data.items():
            if key in relationships:
                relation_dir = relationships[key].direction
                relation_cls = relationships[key].mapper.entity
                if relation_dir == MANYTOMANY:
                    model_data[key] = await create_many_to_many_instances(
                        db_session,
                        relation_cls,
                        value
                    )
                elif relation_dir == ONETOMANY:
                    if relationships[key].uselist:
                        model_data[key] = await create_one_to_many_instances(
                            relation_cls,
                            value
                        )
                    else:
                        # one to one
                        model_data[key] = await create_one_to_one_instance(
                            relation_cls,
                            value
                        )

                elif relation_dir == MANYTOONE:
                    model_data[key] = await create_many_to_one_instance(
                        relation_cls,
                        value
                    )
        entity: ModelType = self.entity(**model_data)
        db_session.add(entity)
        await db_session.flush()
        await self.on_after_create(entity, background_tasks=background_tasks)
        await db_session.commit()
        return entity

    @inject_db_session
    async def crud_create_many(
        self,
        request: Request,
        models: List[CreateSchemaType],
        background_tasks: Optional[BackgroundTasks] = None,
        db_session: Optional[AsyncSession] = Provide()
    ) -> List[ModelType]:
        entities = []
        for model in models:
            entity = await self.crud_create_one(
                request,
                model=model,
                db_session=db_session,
                background_tasks=background_tasks
            )
            entities.append(entity)
        return entities

    @inject_db_session
    async def crud_update_one(
        self,
        request: Request,
        id: ID_TYPE,
        model: UpdateSchemaType,
        db_session: Optional[AsyncSession] = Provide(),
        background_tasks: Optional[BackgroundTasks] = None
    ):
        model_data = model.model_dump(exclude_unset=True)
        relationship_fields = self._guess_should_load_relationship_fields(
            model_data
        )
        joins = functools.reduce(
            lambda x, y: {**x, y: JoinOptionModel(select=True, join=False)},
            relationship_fields,
            {}
        )
        options = self._create_join_options(
            build_join_options_tree(joins),
            request=request,
            from_detail=True
        )
        entity = await self._get(id, db_session=db_session, options=options)
        if not entity:
            raise NotFoundException()
        extra_data = await self.on_before_update(
            entity,
            model,
            background_tasks=background_tasks
        )
        if extra_data:
            model_data.update(extra_data)
        relationships = self.entity.__mapper__.relationships
        for key, value in model_data.items():
            if key in relationships:
                relation_dir = relationships[key].direction
                relation_cls = relationships[key].mapper.entity
                if relation_dir == MANYTOMANY:
                    value = await create_many_to_many_instances(
                        db_session,
                        relation_cls,
                        value
                    )
                elif relation_dir == ONETOMANY:
                    if relationships[key].uselist:
                        value = await create_one_to_many_instances(
                            relation_cls=relation_cls,
                            data=value,
                            old_instances=getattr(entity, key)
                        )
                    else:
                        # one to one
                        value = await create_one_to_one_instance(
                            relation_cls=relation_cls,
                            data=value,
                            old_instance=getattr(entity, key)
                        )
                elif relation_dir == MANYTOONE:
                    value = await create_many_to_one_instance(
                        relation_cls=relation_cls,
                        data=value,
                        old_instance=getattr(entity, key)
                    )
            setattr(entity, key, value)
        db_session.add(entity)
        await db_session.commit()
        await self.on_after_update(entity, background_tasks=background_tasks)
        return entity

    @inject_db_session
    async def crud_update_many(
        self,
        request: Request,
        ids: List[ID_TYPE],
        models: List[UpdateSchemaType],
        background_tasks: Optional[BackgroundTasks] = None,
        db_session: Optional[AsyncSession] = Provide()
    ) -> List[ModelType]:
        if len(ids) != len(models):
            raise Exception("The id and models length do not match")
        entities = []
        for index, model in enumerate(models):
            entity = await self.crud_update_one(
                request,
                id=ids[index],
                model=model,
                db_session=db_session,
                background_tasks=background_tasks
            )
            entities.append(entity)
        return entities

    @inject_db_session
    async def crud_delete_many(
        self,
        request: Request,
        ids: List[ID_TYPE],
        soft_delete: Optional[bool] = False,
        background_tasks: Optional[BackgroundTasks] = None,
        db_session: Optional[AsyncSession] = Provide()
    ) -> List[ModelType]:
        entities = [await self._get(id, db_session) for id in ids]
        for entity in entities:
            if not entity:
                raise NotFoundException()
        await self.on_before_delete(entities, background_tasks=background_tasks)
        if soft_delete:
            await self._soft_delete(ids, db_session=db_session)
        else:
            await self._batch_delete(
                getattr(self.entity, self.primary_key).in_(ids),
                db_session=db_session
            )
        await self.on_after_delete(entities, background_tasks=background_tasks)
        return entities

    def _guess_should_load_relationship_fields(self, model_data: Dict):
        relationships = self.entity.__mapper__.relationships
        relationship_keys = []
        for key in model_data:
            if key in relationships:
                relation_dir = relationships[key].direction
                if relation_dir == MANYTOMANY:
                    relationship_keys.append(key)
                elif relation_dir == ONETOMANY:
                    relationship_keys.append(key)
                elif relation_dir == MANYTOONE:
                    relationship_keys.append(key)
        return relationship_keys

    async def _batch_delete(self, stmt, db_session: AsyncSession):
        if not isinstance(stmt, list):
            stmt = [stmt]
        statement = delete(self.entity).where(*stmt)
        await db_session.execute(statement)
        await db_session.commit()

    async def _get(
        self,
        id: Union[int, str],
        db_session: AsyncSession,
        options: Optional[Sequence[ORMOption]] = None,
    ) -> ModelType:
        return await db_session.get(self.entity, id, options=options, populate_existing=True)

    async def _soft_delete(
        self,
        id_list: List[Union[int, str]],
        db_session: AsyncSession
    ):
        stmt = update(self.entity).where(
            getattr(self.entity, self.primary_key)
            .in_(id_list)).values({
                BetterCrudGlobalConfig.soft_deleted_field_key: datetime.now().replace(microsecond=0)
            })
        await db_session.execute(stmt)
        await db_session.commit()

    async def on_before_create(
        self,
        model: CreateSchemaType,
        background_tasks: Optional[BackgroundTasks] = None
    ) -> Union[Dict[str, Any], None]:
        pass

    async def on_after_create(
        self,
        entity: ModelType,
        background_tasks: BackgroundTasks
    ) -> None:
        pass

    async def on_before_update(
        self,
        entity: ModelType,
        model: UpdateSchemaType,
        background_tasks: BackgroundTasks
    ) -> Union[Dict[str, Any], None]:
        pass

    async def on_after_update(
        self,
        entity: ModelType,
        background_tasks: BackgroundTasks
    ) -> None:
        pass

    async def on_before_delete(
        self,
        entities: List[ModelType],
        background_tasks: BackgroundTasks
    ) -> None:
        pass

    async def on_after_delete(
        self,
        entities: List[ModelType],
        background_tasks: BackgroundTasks
    ) -> None:
        pass

    def build_query_expression(self, field, operator, value):
        if operator == "$eq":
            return field == value
        elif operator == "$ne":
            return field != value
        elif operator == "$gt":
            return field > value
        elif operator == "$gte":
            return field >= value
        elif operator == "$lt":
            return field < value
        elif operator == "$lte":
            return field <= value
        elif operator == "$cont":
            return field.like('%{}%'.format(value))
        elif operator == "$excl":
            return field.notlike('%{}%'.format(value))
        elif operator == "$starts":
            return field.startswith(value)
        elif operator == "$ends":
            return field.endswith(value)
        elif operator == "$notstarts":
            return field.notlike('{}%'.format(value))
        elif operator == "$notends":
            return field.notlike('%{}'.format(value))
        elif operator == "$isnull":
            return field.is_(None)
        elif operator == "$notnull":
            return field.isnot(None)
        elif operator == "$in":
            return field.in_(value.split(","))
        elif operator == "$notin":
            return field.notin_(value.split(","))
        elif operator == "$between":
            return field.between(*value.split(","))
        elif operator == "$notbetween":
            return ~field.between(*value.split(","))
        elif operator == "$length":
            return func.length(field) == int(value)
        elif operator == "$any":
            primary_key = self.get_field_primary_key(field)
            if not primary_key:
                raise NotSupportRelationshipQueryException(operator)
            return field.any(**{primary_key: value})
        elif operator == "$notany":
            primary_key = self.get_field_primary_key(field)
            if not primary_key:
                raise NotSupportRelationshipQueryException(operator)
            return func.not_(field.any(**{primary_key: value}))
        elif operator == "$startsL":
            return field.istartswith(value)
        elif operator == "$endsL":
            return field.iendswith(value)
        elif operator == "$contL":
            return field.ilike('%{}%'.format(value))
        elif operator == "$exclL":
            return field.notilike('%{}%'.format(value))
        elif operator == "$eqL":
            return func.lower(field) == value
        elif operator == "$neL":
            return func.lower(field) != value
        elif operator == "$inL":
            return func.lower(field).in_(value.split(","))
        elif operator == "$notinL":
            return func.lower(field).notin_(value.split(","))
        else:
            raise NotSupportOperatorException(operator)

    def get_model_field(
        self,
        field,
        joins: Optional[JoinOptions] = None
    ):
        field_parts = field.split(".")
        model_field = None
        if len(field_parts) > 1:
            relation_cls = None
            relationships = self.entity.__mapper__.relationships
            for index, field_part in enumerate(field_parts):
                join_key = ".".join(field_parts[0:index+1])
                # query in alias
                if joins and join_key in joins and joins.get(join_key).alias:
                    relation_cls = joins.get(join_key).alias
                    continue
                if relation_cls:
                    model_field = getattr(relation_cls, field_part, None)
                    if index == len(field_parts)-1:
                        break
                relation_cls = relationships[field_part].mapper.entity
                relationships = relation_cls.__mapper__.relationships
        else:
            model_field = getattr(self.entity, field, None)
        if not model_field:
            raise InvalidFieldException(field)
        return model_field

    def get_field_primary_key(self, field):
        try:
            relation_cls = field.mapper.entity
            return relation_cls.__mapper__.primary_key[0].name
        except Exception:
            pass
        return None



from typing import (
    List,
    Optional,
    Any,
    Dict,
    Callable,
    Sequence,
    Literal,
    Union,
    AsyncGenerator,
    TypeVar
)
from typing_extensions import TypedDict
from fastapi import params
from pydantic import BaseModel


RoutesEnumType = Literal[
    "get_many",
    "get_one",
    "create_one",
    "create_many",
    "update_one",
    "update_many",
    "delete_many"
]

BackendType = Literal[
    "sqlalchemy",
    "custom"
]

CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)
ID_TYPE = Union[int, str]
DBSessionFactory = Callable[..., Union[AsyncGenerator[Any, None], Any]]


class RouteSchema(TypedDict):
    name: str
    path: str
    method: Literal[
        "GET",
        "POST",
        "PUT",
        "DELETE"
    ]


class RouteOptionsDict(TypedDict, total=False):
    dependencies: Optional[Sequence[params.Depends]] = None
    summary: Optional[str] = None
    action: Optional[str] = None


class RoutesModelDict(TypedDict, total=False):
    dependencies: Optional[Sequence[params.Depends]] = None
    only: Optional[List[RoutesEnumType]] = None
    exclude: Optional[List[RoutesEnumType]] = None
    get_many: Optional[RouteOptionsDict] = None
    get_one: Optional[RouteOptionsDict] = None
    create_one: Optional[RouteOptionsDict] = None
    create_many: Optional[RouteOptionsDict] = None
    update_one: Optional[RouteOptionsDict] = None
    update_many: Optional[RouteOptionsDict] = None
    delete_many: Optional[RouteOptionsDict] = None


class QueryCriterion(TypedDict, total=False):
    field: str
    value: str
    operator: str


class QuerySortDict(TypedDict):
    field: str
    sort: Literal["ASC", "DESC"]


class PathParamDict(TypedDict):
    field: str
    type: Literal["str", "int"]


class GlobalQueryOptionsDict(TypedDict, total=False):
    soft_delete: Optional[bool] = False
    sort: Optional[List[QuerySortDict]] = None
    allow_include_deleted: Optional[bool] = False


class QueryDelimOptionsDict(TypedDict, total=False):
    delim: Optional[str] = "||"
    delim_str: Optional[str] = ","


class JoinOptionsDict(TypedDict, total=False):
    select: Optional[bool] = True
    join: Optional[bool] = True
    select_only_detail: Optional[bool] = False
    additional_filter_fn: Optional[Callable[[Any], List[Any]]] = None
    alias: Optional[Any] = None


class QueryOptionsDict(TypedDict, total=False):
    joins: Optional[Dict[str, JoinOptionsDict]] = None
    soft_delete: Optional[bool] = None
    allow_include_deleted: Optional[bool] = False
    filter: Union[Optional[Dict], Callable[[Any], Dict]] = None
    sort: Optional[List[QuerySortDict]] = None


class AuthModelDict(TypedDict, total=False):
    filter: Optional[Callable[[Any], Dict]]
    persist: Optional[Callable[[Any], Dict]]


class DtoModelDict(TypedDict, total=False):
    create: Optional[Any] = None
    update: Optional[Any] = None


class SerializeModelDict(TypedDict, total=False):
    base: Any
    get_many: Optional[Any] = None
    get_one: Optional[Any] = None
    create_one: Optional[Any] = None
    create_many: Optional[Any] = None
    update_one: Optional[Any] = None
    delete_many: Optional[Any] = None


class SqlalchemyBackendDict(TypedDict):
    db_session: DBSessionFactory


class BackendConfigDict(TypedDict):
    backend: BackendType
    sqlalchemy: SqlalchemyBackendDict



from typing import Dict, List, Union, Optional, Any
import inspect
from contextlib import asynccontextmanager
from sqlalchemy.ext.asyncio import AsyncSession
from ...helper import find, update_entity_attr
from ...config import BetterCrudGlobalConfig


class Provide:
    pass


async def create_many_to_many_instances(
    session: AsyncSession,
    relation_cls,
    data: List[Union[Dict, int]]
) -> List[Any]:
    primary_key = relation_cls.__mapper__.primary_key[0].name
    if len(data) > 0 and isinstance(data[0], dict):
        data = [elem[primary_key] for elem in data]
    instances = [
        await session.get(relation_cls, primary_value)
        for primary_value in data
    ]
    return instances


async def create_one_to_many_instances(
    relation_cls: Any,
    data: List[Dict],
    old_instances: Optional[Any] = None
) -> List[Any]:
    primary_key = relation_cls.__mapper__.primary_key[0].name
    instances = []
    for item_data in data:
        if old_instances and primary_key in item_data:
            instance = find(
                old_instances,
                lambda x: getattr(x, primary_key) == item_data.get(primary_key)
            )
            if instance:
                update_entity_attr(instance, item_data)
                instances.append(instance)
                continue
        instances.append(relation_cls(**item_data))
    return instances


async def create_many_to_one_instance(
    relation_cls: Any,
    data: Dict,
    old_instance: Optional[Any] = None
):
    if old_instance is None:
        return relation_cls(**data)
    update_entity_attr(old_instance, data)
    return old_instance


async def create_one_to_one_instance(
    relation_cls: Any,
    data: Dict,
    old_instance: Optional[Any] = None
):
    if old_instance is None:
        return relation_cls(**data)
    update_entity_attr(old_instance, data)
    return old_instance


def inject_db_session(f):
    sig = inspect.signature(f)

    async def wrapper(*args, **kwargs):
        for param in sig.parameters.values():
            if isinstance(param.default, Provide):
                if kwargs.get(param.name) is None:
                    sqlalchemy_config = BetterCrudGlobalConfig.backend_config.sqlalchemy
                    if inspect.isasyncgenfunction(sqlalchemy_config.db_session):
                        DBSession = asynccontextmanager(
                            sqlalchemy_config.db_session)
                        async with DBSession() as db_session:
                            kwargs[param.name] = db_session
                            return await f(*args, **kwargs)
                    else:
                        kwargs[param.name] = sqlalchemy_config.db_session()
                        return await f(*args, **kwargs)
                else:
                    return await f(*args, **kwargs)
    return wrapper



from typing import Dict
from .service.abstract import AbstractCrudService
backend_cls_map: Dict[str, AbstractCrudService] = {}


def register_backend(name: str):
    def decorator(backend_cls):
        backend_cls_map[name] = backend_cls
        return backend_cls
    return decorator


def get_backend(name: str):
    return backend_cls_map.get(name)
